from .__version__ import __version__

from swiftemulator.backend.model_specification import ModelSpecification
from swiftemulator.backend.model_parameters import ModelParameters
from swiftemulator.backend.model_values import ModelValues



"""
Tests for objects in ``swift-emulator/backend/model_specification.py``
"""

from swiftemulator.backend.model_specification import ModelSpecification
import pytest


def test_basic_creation():
    """
    Basic creation of the model specification.
    """

    model_spec = ModelSpecification(
        number_of_parameters=3,
        parameter_names=["Hello", "World", "Foo"],
        parameter_limits=[[0.0, 1.0], [0.0, 1.0], [0.2, 0.3]],
    )

    assert model_spec.parameter_names == model_spec.parameter_printable_names

    return


def test_failure():
    """
    Test the failure when parameter limits are not included.
    """

    with pytest.raises(AttributeError):
        _ = ModelSpecification(
            number_of_parameters=3,
            parameter_names=["Hello", "World", "Foo"],
            parameter_limits=[[0.0, 1.0], [0.0, 1.0]],
        )

    return



"""
Visualisation functions for comparison datasets.

Allows you to project a plausibility region for each
parameter cross-correlation.
"""

from swiftemulator.backend.model_parameters import ModelParameters
from swiftemulator.backend.model_specification import ModelSpecification
from swiftemulator.backend.model_values import ModelValues

from typing import Dict, Hashable, Tuple, Iterable, Optional
from matplotlib.colors import Normalize
from scipy.stats import binned_statistic_2d

try:
    from swiftsimio.visualisation.projection import scatter

    swiftsimio_available = True
except (ImportError, ModuleNotFoundError):
    swiftsimio_available = False

import matplotlib.pyplot as plt
import numpy as np


def visualise_penalties_mean(
    model_specification: ModelSpecification,
    model_parameters: ModelParameters,
    penalties: Dict[Hashable, float],
    norm: Normalize = Normalize(vmin=0.2, vmax=0.7, clip=True),
    remove_ticks: bool = True,
    figsize: Optional[Tuple[float]] = None,
    use_parameters: Optional[Iterable[str]] = None,
    use_colorbar: Optional[bool] = False,
    highlight_model: Optional[Hashable] = None,
) -> Tuple[plt.Figure, Iterable[plt.Axes]]:
    """
    Visualises the penalties using SPH smoothing for each
    individual model point.

    Parameters
    ----------

    model_specification: ModelSpecification
        The appropriate model specification. Used for the limits
        of the figure.

    model_parameters: ModelParameters
        Parameters of the model, with the appropriate unique IDs.

    penalties: Dict[Hashable, float]
        Penalties for all parameters in ``model_parameters``, with
        the key in this dictionary being the unique IDs.

    norm: Normalize, optional
        A ``matplotlib`` normalisation object. By default this uses
        ``vmin=0.2`` and ``vmax=0.7``.

    remove_ticks: bool, optional
        Remove the axes ticks? This is recommended, as the plot can
        become very cluttered if you don't do this. Default: ``True``.

    figsize: Tuple[float], optional
        The figure size to use. Defaults to 7 inches by 7 inches, the
        size for a ``figure*`` in the MNRAS template.

    use_parameters: Iterable[str], optional
        The parameters to include in the figure. If not provided, all
        parameters in the ``model_specification`` are used.

    use_colorbar: Bool, optional
        Include a colorbar? Default: False

    highlight_model: Hashable, optional
        The model unique ID to highlight. If not provided, no model is
        highlighted.

    Returns
    -------

    fig: Figure
        The figure object.

    axes: np.ndarray[Axes]
        The individual axes.

    Notes
    -----

    You can either change how the figure looks by using the figure
    and axes objects that are returned, or by modifying the
    ``matplotlib`` stylesheet you are currently using.
    """

    if use_parameters is None:
        use_parameters = model_specification.parameter_names

    if figsize is None:
        if use_colorbar:
            figsize = (7.0, 7.4)
        else:
            figsize = (7.0, 7.0)

    parameter_indices = [
        model_specification.parameter_names.index(x) for x in use_parameters
    ]

    number_of_parameters = len(use_parameters)
    grid_size = number_of_parameters

    fig, axes_grid = plt.subplots(
        grid_size,
        grid_size,
        figsize=figsize,
        squeeze=True,
        sharex="col",
        sharey="row",
    )

    visualisation_size = 2.0 / np.sqrt(len(model_parameters))
    simulation_ordering = list(model_parameters.keys())

    if highlight_model is not None:
        highlight_index = simulation_ordering.index(highlight_model)

    # Build temporary 1D arrays of parameters/offsets in correct ordering
    ordered_penalties = np.array([penalties[x] for x in simulation_ordering])

    limits = model_specification.parameter_limits
    # Parameters must be re-scaled to the range [0,1] for smoothed projection.
    ordered_parameters = [
        (
            np.array(
                [
                    model_parameters.model_parameters[x][parameter]
                    for x in simulation_ordering
                ]
            )
            - limits[index][0]
        )
        / (limits[index][1] - limits[index][0])
        for index, parameter in zip(parameter_indices, use_parameters)
    ]

    smoothing_lengths = np.ones_like(ordered_penalties) * visualisation_size

    for parameter_x, axes_column in zip(parameter_indices, axes_grid):
        for parameter_y, ax in zip(parameter_indices, axes_column):
            limits_x = model_specification.parameter_limits[parameter_x]
            limits_y = model_specification.parameter_limits[parameter_y]
            name_x = model_specification.parameter_printable_names[parameter_x]
            name_y = model_specification.parameter_printable_names[parameter_y]

            is_center_line = parameter_x == parameter_y
            do_not_plot = is_center_line and remove_ticks

            if not do_not_plot:
                if swiftsimio_available:
                    norm_grid = scatter(
                        x=ordered_parameters[parameter_x],
                        y=ordered_parameters[parameter_y],
                        m=np.ones_like(ordered_penalties),
                        h=smoothing_lengths,
                        res=512,
                    )

                    weighted_grid = scatter(
                        x=ordered_parameters[parameter_x],
                        y=ordered_parameters[parameter_y],
                        m=ordered_penalties,
                        h=smoothing_lengths,
                        res=512,
                    )

                    norm_grid[norm_grid == 0.0] = 1.0

                    ratio_grid = weighted_grid / norm_grid
                else:
                    norm_grid, _, _, _ = binned_statistic_2d(
                        x=ordered_parameters[parameter_x],
                        y=ordered_parameters[parameter_y],
                        values=ordered_penalties,
                        statistic="mean",
                        bins=16,
                    )

                im = ax.imshow(
                    ratio_grid.T,
                    extent=limits_x + limits_y,
                    origin="lower",
                    norm=norm,
                    rasterized=True,
                )

                if highlight_model is not None:
                    highlight_x = ordered_parameters[parameter_x][highlight_index]
                    highlight_y = ordered_parameters[parameter_y][highlight_index]

                    # Need to re-scale from 0->1 to 'real' space

                    highlight_x *= limits_x[1] - limits_x[0]
                    highlight_x += limits_x[0]

                    highlight_y *= limits_y[1] - limits_y[0]
                    highlight_y += limits_y[0]

                    ax.scatter(
                        highlight_x,
                        highlight_y,
                        color="white",
                        edgecolor="black",
                    )

                ax.set_xlim(*limits_x)
                ax.set_ylim(*limits_y)

            if remove_ticks:
                ax.tick_params(
                    axis="both",
                    which="both",
                    bottom=False,
                    left=False,
                    right=False,
                    top=False,
                    labelbottom=False,
                    labelleft=False,
                    labelright=False,
                    labeltop=False,
                )
                if is_center_line:
                    ax.text(
                        0.5,
                        0.5,
                        f"{limits_x[0]:3.3f} <\n{name_x}\n< {limits_x[1]:3.3f}",
                        transform=ax.transAxes,
                        ha="center",
                        va="center",
                    )

            else:
                ax.set_xlabel(name_x)
                ax.set_ylabel(name_y)

            # Set square in data reference frame
            ax.set_aspect(1.0 / ax.get_data_ratio())

    if use_colorbar:
        fig.colorbar(
            im,
            ax=axes_grid.ravel().tolist(),
            orientation="horizontal",
            label="Mean penalty along line of sight",
        )

    for a in axes_grid[:-1, :].flat:
        a.set_xlabel(None)
    for a in axes_grid[:, 1:].flat:
        a.set_ylabel(None)

    # As of matplotlib 3.3.4, with a large number of sub-plots this hangs...
    if grid_size > 4:
        fig.constrained_layout = False
        fig.subplots_adjust(0, 0, 1, 1, 0.005, 0.005)

    return fig, ax


def visualise_penalties_generic_statistic(
    model_specification: ModelSpecification,
    model_parameters: ModelParameters,
    penalties: Dict[Hashable, float],
    statistic: Optional[str] = None,
    norm: Normalize = Normalize(vmin=0.2, vmax=0.7, clip=True),
    remove_ticks: bool = True,
    figsize: Optional[Tuple[float]] = None,
    use_parameters: Optional[Iterable[str]] = None,
    use_colorbar: Optional[bool] = False,
    highlight_model: Optional[Hashable] = None,
) -> Tuple[plt.Figure, Iterable[plt.Axes]]:
    """
    Visualises the penalties using basic binning.

    Parameters
    ----------

    model_specification: ModelSpecification
        The appropriate model specification. Used for the limits
        of the figure.

    model_parameters: ModelParameters
        Parameters of the model, with the appropriate unique IDs.

    penalties: Dict[Hashable, float]
        Penalties for all parameters in ``model_parameters``, with
        the key in this dictionary being the unique IDs.

    statistic: str, optional
        The statistic that you would like to compute. Allowed values
        are the same as for ``scipy.stats.binned_statistic_2d``.
        Defaults to ``mean``.

    norm: Normalize, optional
        A ``matplotlib`` normalisation object. By default this uses
        ``vmin=0.2`` and ``vmax=0.7``.

    remove_ticks: bool, optional
        Remove the axes ticks? This is recommended, as the plot can
        become very cluttered if you don't do this. Default: ``True``.

    figsize: Tuple[float], optional
        The figure size to use. Defaults to 7 inches by 7 inches, the
        size for a ``figure*`` in the MNRAS template.

    use_parameters: Iterable[str], optional
        The parameters to include in the figure. If not provided, all
        parameters in the ``model_specification`` are used.

    use_colorbar: Bool, optional
        Include a colorbar? Default: False.

    highlight_model: Hashable, optional
        The model unique ID to highlight. If not provided, no model is
        highlighted.


    Returns
    -------

    fig: Figure
        The figure object.

    axes: np.ndarray[Axes]
        The individual axes.

    Notes
    -----

    You can either change how the figure looks by using the figure
    and axes objects that are returned, or by modifying the
    ``matplotlib`` stylesheet you are currently using.
    """

    if use_parameters is None:
        use_parameters = model_specification.parameter_names

    if figsize is None:
        if use_colorbar:
            figsize = (7.0, 7.4)
        else:
            figsize = (7.0, 7.0)

    parameter_indices = [
        model_specification.parameter_names.index(x) for x in use_parameters
    ]

    number_of_parameters = len(use_parameters)
    grid_size = number_of_parameters

    fig, axes_grid = plt.subplots(
        grid_size,
        grid_size,
        figsize=figsize,
        squeeze=True,
        sharex="col",
        sharey="row",
    )

    visualisation_size = 4.0 / np.sqrt(len(model_parameters))
    simulation_ordering = list(model_parameters.keys())

    if highlight_model is not None:
        highlight_index = simulation_ordering.index(highlight_model)

    # Build temporary 1D arrays of parameters/offsets in correct ordering
    ordered_penalties = np.array([penalties[x] for x in simulation_ordering])

    limits = model_specification.parameter_limits

    ordered_parameters = [
        np.array(
            [
                model_parameters.model_parameters[x][parameter]
                for x in simulation_ordering
            ]
        )
        for index, parameter in zip(parameter_indices, use_parameters)
    ]

    bins = int(round(1.0 / visualisation_size))

    statistic = statistic if statistic is not None else "mean"

    # JB: I am 100% confident in this loop and that we are looping
    # over the correct axes. Do not change this loop.
    for parameter_y, axes_column in zip(parameter_indices, axes_grid):
        for parameter_x, ax in zip(parameter_indices, axes_column):
            limits_x = model_specification.parameter_limits[parameter_x]
            limits_y = model_specification.parameter_limits[parameter_y]
            name_x = model_specification.parameter_printable_names[parameter_x]
            name_y = model_specification.parameter_printable_names[parameter_y]

            is_center_line = parameter_x == parameter_y
            do_not_plot = is_center_line and remove_ticks

            if not do_not_plot:
                grid, xs, ys, _ = binned_statistic_2d(
                    x=ordered_parameters[parameter_x],
                    y=ordered_parameters[parameter_y],
                    values=ordered_penalties,
                    statistic=statistic,
                    bins=bins,
                )

                im = ax.pcolormesh(
                    xs,
                    ys,
                    grid.T,
                    norm=norm,
                    rasterized=True,
                )

                # Uncomment me if you don't believe the comment above
                # ax.text(0.5, 0.5, f"x={name_x}\ny={name_y}", transform=ax.transAxes, ha="center", va="center", color="white")

                if highlight_model is not None:
                    highlight_x = ordered_parameters[parameter_x][highlight_index]
                    highlight_y = ordered_parameters[parameter_y][highlight_index]

                    ax.scatter(
                        highlight_x,
                        highlight_y,
                        color="white",
                        edgecolor="black",
                    )

                ax.set_xlim(*limits_x)
                ax.set_ylim(*limits_y)

            if remove_ticks:
                ax.tick_params(
                    axis="both",
                    which="both",
                    bottom=False,
                    left=False,
                    right=False,
                    top=False,
                    labelbottom=False,
                    labelleft=False,
                    labelright=False,
                    labeltop=False,
                )

                if is_center_line:
                    ax.text(
                        0.5,
                        0.5,
                        f"{limits_x[0]:3.3f} <\n{name_x}\n< {limits_x[1]:3.3f}",
                        transform=ax.transAxes,
                        ha="center",
                        va="center",
                    )
            else:
                ax.set_xlabel(name_x)
                ax.set_ylabel(name_y)

            # Set square in data reference frame
            ax.set_aspect(1.0 / ax.get_data_ratio())

    if use_colorbar:
        fig.colorbar(
            im,
            ax=axes_grid.ravel().tolist(),
            orientation="horizontal",
            label=f"{statistic.capitalize()} penalty along line of sight",
        )

    for a in axes_grid[:-1, :].flat:
        a.set_xlabel(None)
    for a in axes_grid[:, 1:].flat:
        a.set_ylabel(None)

    # As of matplotlib 3.3.4, with a large number of sub-plots this hangs...
    if grid_size > 4:
        fig.constrained_layout = False
        fig.subplots_adjust(0, 0, 1, 1, 0.005, 0.005)

    return fig, ax



"""
Model specification for the emulator.

Contains the :class:``ModelSpecification`` class,
used to input the parameter names and limits.
"""

import attr
from typing import List, Optional


@attr.s
class ModelSpecification(object):
    """
    Base specification for the model. Contains information about
    the names of parameters and their ranges.

    Parameters
    ----------

    number_of_parameters, int
        Total number of variable parameters in your model.

    parameter_names, List[str]
        The names of parameters in your model; these will be
        used to access the parameters in :class:``ModelParameters``.

    parameter_limits, List[List[float]]
        The lower and upper limit of the input parameters. Should be
        the same length as ``parameter_names``, but each item is a list
        of length two, with a lower and upper bound. For example, in
        a two parameter model ``[[0.0, 1.0], [8.3, 9.3]]`` would mean
        that the first parameter would vary between 0.0 and 1.0, with
        the second parameter varying between 8.3 and 9.3.

    parameter_printable_names, List[str], optional
        Optional 'fancy' names for your parameters. These strings will
        be used on any figures generated through swift-emulator. Can
        include LaTeX formatting as in ``matplotlib``.

    Raises
    ------

    AttributeError
        When the number of paremeters in all of the required attributes
        are not equal (e.g. a different number of names has been provided
        compared to the number of limits).

    """

    number_of_parameters: int = attr.ib(validator=attr.validators.instance_of(int))
    parameter_names: List[str] = attr.ib()
    parameter_limits: List[List[float]] = attr.ib()
    parameter_printable_names: Optional[List[str]] = attr.ib(default=None)

    @parameter_limits.validator
    def _check_parameter_limits(self, attribute, value):
        if not len(self.parameter_names) == len(value):
            raise AttributeError(
                "Parameter limits and parameter names not of same length."
            )

    def __attrs_post_init__(self):
        if not self.number_of_parameters == len(self.parameter_names):
            raise AttributeError(
                "Number of parameters does not match the length of parameter names"
            )
        if self.parameter_printable_names is None:
            self.parameter_printable_names = self.parameter_names

    @property
    def salib_problem(self):
        """
        Generates the ``SALib`` ``problem`` dictionary.
        """

        problem = {
            "num_vars": self.number_of_parameters,
            "names": self.parameter_names,
            "bounds": self.parameter_limits,
        }

        return problem



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("swiftemulator/__version__.py", "r") as fh:
    exec_output = {}
    exec(fh.read(), exec_output)
    __version__ = exec_output["__version__"]

setuptools.setup(
    name="swiftemulator",
    version=__version__,
    description="Gaussian process emulator for creating synthetic model data across high dimensional parameter spaces, initially developed for use with the SWIFT simulation code.",
    url="https://github.com/SWIFTSIM/emulator",
    author="Josh Borrow",
    author_email="josh@joshborrow.com",
    packages=setuptools.find_packages(),
    long_description=long_description,
    long_description_content_type="text/markdown",
    zip_safe=False,
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "numpy",
        "unyt",
        "attrs",
        "george",
        "SALib",
        "scikit-learn",
        "corner",
        "velociraptor",
        "pyyaml",
        "pyDOE",
        "tqdm",
        "emcee",
    ],
    include_package_data=True,
)



"""
Base for all emulators. This :class:`BaseEmulator` class is a specification
for how all emulators should be implemented, but does not implement a
specific model itself (all methods raise ``NotImplementedError`` by default).
Controls on the behaviour of the emulator are passed as initialiser
arguments, and the data is passed as arguments to fit_model.
"""

import attr
import copy
import numpy as np

from typing import Hashable, List, Optional, Dict

from swiftemulator.backend.model_parameters import ModelParameters
from swiftemulator.backend.model_specification import ModelSpecification
from swiftemulator.backend.model_values import ModelValues


@attr.s
class BaseEmulator(object):
    """
    Base emulator for training models. Initialisation parameters
    are used in ``fit_model`` to specify additional parameters
    to the model.
    """

    ordering: Optional[List[Hashable]] = None
    parameter_order: Optional[List[str]] = None

    model_specification: Optional[ModelSpecification] = None
    model_parameters: Optional[ModelParameters] = None
    model_values: Optional[ModelValues] = None

    independent_variables: Optional[np.array] = None
    dependent_variables: Optional[np.array] = None
    dependent_variable_errors: Optional[np.array] = None

    emulator = None

    def _build_arrays(
        self,
        model_specification: ModelSpecification,
        model_parameters: ModelParameters,
        model_values: ModelValues,
    ):
        """
        Builds the internal arrays based on the given model specification,
        parameters, and values.
        """

        self.model_specification = model_specification
        self.model_parameters = model_parameters
        self.model_values = model_values

        self.independent_variables = None
        self.dependent_variables = None
        self.dependent_variable_errors = None

        raise NotImplementedError

    def fit_model(
        self,
        model_specification: ModelSpecification,
        model_parameters: ModelParameters,
        model_values: ModelValues,
    ):
        """
        Fits a model to the independent and dependent variables given by
        the model spec, parameters, and values.
        """

        self._build_arrays()

        self.emulator = None

        raise NotImplementedError

    def predict_values(
        self, independent: np.array, model_parameters: Dict[str, float]
    ) -> np.array:
        """
        Predict values and the associated variance from the trained emulator contained
        within this object.

        Parameters
        ----------

        independent: np.array
            Independent continuous variables to evaluate the emulator
            at. If the emulator is discrete, these are only allowed to be
            the discrete independent variables that the emulator was trained at
            (disregarding the additional 'independent' model parameters, below.)

        model_parameters: Dict[str, float]
            The point in model parameter space to create predicted
            values at.

        Returns
        -------

        dependent_predictions: np.array
            Array of predictions, if the emulator is a function f, these
            are the predicted values of f(independent) evaluted at the position
            of the input ``model_parameters``.

        dependent_prediction_errors: np.array
            Errors on the model predictions. For models where the errors are
            unconstrained, this is an array of zeroes.

        Raises
        ------

        AttributeError
            When the model has not been trained before trying to make a
            prediction, or when attempting to evaluate the model at
            disallowed independent variables.
        """

        if self.emulator is None:
            raise AttributeError(
                "Please train the emulator with fit_model before attempting "
                "to make predictions."
            )

        raise NotImplementedError

    def predict_values_no_error(
        self, independent: np.array, model_parameters: Dict[str, float]
    ) -> np.array:
        """
        Predict values from the trained emulator contained within this object.
        In cases where the error estimates are not required, this method is
        significantly faster than predict_values().

        Parameters
        ----------

        independent, np.array
            Independent continuous variables to evaluate the emulator
            at. If the emulator is discrete, these are only allowed to be
            the discrete independent variables that the emulator was trained at
            (disregarding the additional 'independent' model parameters, below.)

        model_parameters: Dict[str, float]
            The point in model parameter space to create predicted
            values at.

        Returns
        -------

        dependent_predictions, np.array
            Array of predictions, if the emulator is a function f, these
            are the predicted values of f(independent) evaluted at the position
            of the input ``model_parameters``.

        Raises
        ------

        AttributeError
            When the model has not been trained before trying to make a
            prediction, or when attempting to evaluate the model at
            disallowed independent variables.
        """

        if self.emulator is None:
            raise AttributeError(
                "Please train the emulator with fit_model before attempting "
                "to make predictions."
            )

        raise NotImplementedError

    def interactive_plot(
        self,
        x: np.array,
        initial_params: Dict[str, float] = {},
        xlabel: str = "",
        ylabel: str = "",
        x_data: np.array = None,
        y_data: np.array = None,
    ):
        """
        Generates an interactive plot which displays the emulator predictions.
        If initial_params should contain the initial parameter values to make a
        prediction for. If initial_params is not passed the midpoint of each of
        the parameter values will be used instead. If no reference data is
        passed to be overplotted then the plot will display a line which
        corresponds to the predictions for the initial parameter values.

        Parameters
        ----------

        x: np.array
            Array of data for which the emulator should make predictions.

        initial_params: Dict[str, float], optional
            What parameters values to plot the predicition for initally.
            If missing the midpoint of each parameter range will be used.

        xlabel: str, optional
            Label for horizontal axis on the resultant figure.

        ylabel: str, optional
            Label for vertical axis on the resultant figure.

        x_data: np.array, optional
            Array containing x-values of reference data to plot.

        y_data: np.array, optional
            Array containing y-values of reference data to plot.
            Must be the same shape as x_data
        """

        import matplotlib.pyplot as plt
        from matplotlib.widgets import Slider

        fig, ax = plt.subplots()
        model_specification = self.model_specification
        sliders = []
        n_param = model_specification.number_of_parameters
        fig.subplots_adjust(bottom=0.12 + n_param * 0.1)
        for i in range(n_param):
            # Extracting information needed for slider
            name = model_specification.parameter_names[i]
            lo_lim = sorted(model_specification.parameter_limits[i])[0]
            hi_lim = sorted(model_specification.parameter_limits[i])[1]
            if not name in initial_params:
                initial_params[name] = (lo_lim + hi_lim) / 2

            # Adding slider
            printable_name = name
            if model_specification.parameter_printable_names:
                printable_name = model_specification.parameter_printable_names[i]
            slider_ax = fig.add_axes([0.35, i * 0.1, 0.3, 0.1])
            slider = Slider(
                ax=slider_ax,
                label=printable_name,
                valmin=lo_lim,
                valmax=hi_lim,
                valinit=initial_params[name],
            )
            sliders.append(slider)

        # Plotting lines and reference data
        pred, pred_var = self.predict_values(x, initial_params)
        if (x_data is None) or (y_data is None):
            ax.plot(x, pred, "k--")
        else:
            ax.plot(x_data, y_data, "k.")
        (line,) = ax.plot(x, pred)
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)

        # Define and enable update function
        def update(val):
            params = {
                model_specification.parameter_names[i]: sliders[i].val
                for i in range(n_param)
            }
            pred, pred_var = self.predict_values(x, params)
            line.set_ydata(pred)

        for slider in sliders:
            slider.on_changed(update)

        plt.show()
        plt.close()



"""
Gaussian Process Emulator using an emulator for each bins
"""

import attr
import copy
import numpy as np
import george

from typing import Hashable, List, Optional, Dict

from swiftemulator.emulators.base import BaseEmulator

from swiftemulator.backend.model_parameters import ModelParameters
from swiftemulator.backend.model_specification import ModelSpecification
from swiftemulator.backend.model_values import ModelValues
from swiftemulator.emulators.gaussian_process import GaussianProcessEmulator
from swiftemulator.mean_models import MeanModel

from scipy.optimize import minimize


@attr.s
class GaussianProcessEmulatorBins(BaseEmulator):
    """
    Generator for emulators for individual scaling relations.
    Uses a GP for each seperate bin.

    Parameters
    ----------

    kernel, george.kernels.Kernel, optional
        The ``george`` kernel to use. The GPE here uses a copy
        of this instance. By default, this is the
        ``ExpSquaredKernel`` in George

    mean_model, MeanModel, optional
        A mean model conforming to the ``swiftemulator`` mean model
        protocol (several pre-made models are available in the
        :mod:`swiftemulator.mean_models` module).
    """

    kernel: Optional[george.kernels.Kernel] = attr.ib(default=None)
    mean_model: Optional[MeanModel] = attr.ib(default=None)

    model_specification: Optional[ModelSpecification] = None
    model_parameters: Optional[ModelParameters] = None
    model_values: Optional[ModelValues] = None

    ordering: Optional[List[Hashable]] = None
    parameter_order: Optional[List[str]] = None

    independent_variables: Optional[np.array] = None
    dependent_variables: Optional[np.array] = None
    dependent_variable_errors: Optional[np.array] = None

    n_bins: int = None
    bin_model_values: List[Dict[str, np.array]] = None
    bin_centers: List[float] = None
    bin_gaussian_process: List[george.GP] = None

    def _build_arrays(
        self,
        model_specification: ModelSpecification,
        model_parameters: ModelParameters,
        model_values: ModelValues,
    ):
        """
        Builds the arrays for passing to `george`. As we aim to build
        a emulator for each dependent bin, this creates a dictionary containing
        the dependent, dependent errors and the variables beloning to each bin
        in a format the `george` likes. It also links this to the bin centers,
        which are stored as a separate dictionary.

        Parameters
        ----------

        model_specification: ModelSpecification
            Full instance of the model specification.

        model_parameters: ModelParameters
            Full instance of the model parameters.

        model_values: ModelValues
            Full instance of the model values describing
            this individual scaling relation.
        """

        self.model_specification = model_specification
        self.model_parameters = model_parameters
        self.model_values = model_values

        self.parameter_order = self.model_specification.parameter_names
        number_of_model_parameters = self.model_specification.number_of_parameters
        unique_identifiers = list(model_values.model_values.keys())

        bin_centers = np.unique(
            [
                item
                for uid in unique_identifiers
                for item in model_values.model_values[uid]["independent"]
            ]
        )

        self.n_bins = len(bin_centers)
        self.bin_centers = bin_centers

        bin_model_values = []

        for bin_center in bin_centers:
            bin_variables = []
            bin_dependent = []
            bin_dependent_errors = []

            for unique_identifier in unique_identifiers:
                uniq_model_values = model_values.model_values[unique_identifier]
                bin_independent_uniq = uniq_model_values["independent"]

                # Which one of our bins corresponds to the bin of interest?
                # bin_mask only has one (or zero) True values.
                bin_mask = bin_independent_uniq == bin_center

                if not bin_mask.any():
                    continue

                bin_dependent_uniq = uniq_model_values["dependent"][bin_mask][0]
                bin_dependent_errors_uniq = uniq_model_values.get(
                    "dependent_error", np.zeros(len(bin_independent_uniq))
                )[bin_mask][0]
                bin_independent_uniq = bin_independent_uniq[bin_mask][0]

                bin_variables_uniq = [
                    model_parameters.model_parameters[unique_identifier][parameter]
                    for parameter in self.parameter_order
                ]

                bin_dependent.append(bin_dependent_uniq)
                bin_dependent_errors.append(bin_dependent_errors_uniq)
                bin_variables.append(bin_variables_uniq)

                if np.ndim(bin_dependent_errors) != 1:
                    raise AttributeError(
                        "Multiple dimensional errors are not currently supported in GPE mode"
                    )

            bin_variables = np.array(bin_variables).reshape(
                (int(len(bin_dependent)), number_of_model_parameters)
            )

            bin_model_values.append(
                {
                    "independent": bin_variables,
                    "dependent": np.array(bin_dependent).flatten(),
                    "dependent_errors": np.array(bin_dependent_errors).flatten(),
                }
            )

        self.bin_model_values = bin_model_values

        return

    def fit_model(
        self,
        model_specification: ModelSpecification,
        model_parameters: ModelParameters,
        model_values: ModelValues,
    ):
        """
        Fits the gaussian process model, as determined by the
        initialiser variables of the class (i.e. the kernel and
        the mean model).

        Parameters
        ----------

        model_specification: ModelSpecification
            Full instance of the model specification.

        model_parameters: ModelParameters
            Full instance of the model parameters.

        model_values: ModelValues
            Full instance of the model values describing
            this individual scaling relation.

        Notes
        -----

        This method uses copies of the internal kernel and mean model
        objects, as those objects contain slightly unhelpful state information.
        """

        if self.bin_model_values is None:
            # Creates independent_variables, dependent_variables.
            self._build_arrays(
                model_specification=model_specification,
                model_parameters=model_parameters,
                model_values=model_values,
            )

        unique_bin_identifiers = list(range(self.n_bins))
        bin_model_values = self.bin_model_values

        bin_gaussian_process = []

        for bin_index in unique_bin_identifiers:
            independent_variables = bin_model_values[bin_index]["independent"]
            dependent_variables = bin_model_values[bin_index]["dependent"]
            dependent_variable_errors = bin_model_values[bin_index]["dependent_errors"]

            if self.kernel is None:
                number_of_kernel_dimensions = (
                    self.model_specification.number_of_parameters
                )

                self.kernel = 1**2 * george.kernels.ExpSquaredKernel(
                    np.ones(number_of_kernel_dimensions),
                    ndim=number_of_kernel_dimensions,
                )

            if self.mean_model is not None:
                self.mean_model.train(
                    independent=independent_variables,
                    dependent=dependent_variables,
                )

                gaussian_process = george.GP(
                    copy.deepcopy(self.kernel),
                    fit_kernel=True,
                    mean=self.mean_model.george_model,
                    fit_mean=False,
                )
            else:
                gaussian_process = george.GP(
                    copy.deepcopy(self.kernel),
                )

            # TODO: Figure out how to include non-symmetric errors.
            gaussian_process.compute(
                x=independent_variables,
                yerr=dependent_variable_errors,
            )

            def negative_log_likelihood(p):
                gaussian_process.set_parameter_vector(p)
                return -gaussian_process.log_likelihood(dependent_variables)

            def grad_negative_log_likelihood(p):
                gaussian_process.set_parameter_vector(p)
                return -gaussian_process.grad_log_likelihood(dependent_variables)

            # Optimize the hyperparameter values in the emulator
            result = minimize(
                fun=negative_log_likelihood,
                x0=gaussian_process.get_parameter_vector(),
                jac=grad_negative_log_likelihood,
            )

            # Load in the optimal hyperparameters
            gaussian_process.set_parameter_vector(result.x)

            bin_gaussian_process.append(gaussian_process)

        self.bin_gaussian_process = bin_gaussian_process

        return

    def predict_values(
        self, independent: np.array, model_parameters: Dict[str, float]
    ) -> np.array:
        """
        Predict values and the associated variance from the trained emulator contained
        within this object.

        Parameters
        ----------

        independent, np.array
            Independent continuous variables to evaluate the emulator
            at. If the emulator is discrete, these are only allowed to be
            the discrete independent variables that the emulator was trained at
            (disregarding the additional 'independent' model parameters, below).
            These can be found in this object in the ``bin_centers`` attribute.

        model_parameters: Dict[str, float]
            The point in model parameter space to create predicted
            values at.

        Returns
        -------

        dependent_predictions, np.array
            Array of predictions, if the emulator is a function f, these
            are the predicted values of f(independent) evaluted at the position
            of the input ``model_parameters``.

        dependent_prediction_errors, np.array
            Errors on the model predictions. For models where the errors are
            unconstrained, this is an array of zeroes.

        Raises
        ------

        AttributeError
            When the model has not been trained before trying to make a
            prediction, or when attempting to evaluate the model at
            disallowed independent variables.
        """

        if self.bin_gaussian_process is None:
            raise AttributeError(
                "Please train the emulator with fit_model before attempting "
                "to make predictions."
            )

        # First calculate which indices in bin_centers (and hence
        # bin_gaussian_processes) correspond to the requested ``independent``
        # variables.

        array_centers = np.array(self.bin_centers)
        gpe_ordering = []

        for requested_independent_variable in independent:
            try:
                gpe_ordering.append(
                    np.where(array_centers == requested_independent_variable)[0][0]
                )
            except IndexError:
                raise AttributeError(
                    f"Requested independent variable {independent} not valid, ",
                    f"this instance of GPE Bins is only valid at {array_centers}.",
                )

        model_parameter_array = np.array(
            [model_parameters[parameter] for parameter in self.parameter_order]
        )

        # George must predict a value for more than one point at a time, so
        # generate two fake points either side of the one of interest.
        model_parameter_array_sample = np.append(
            0.98 * model_parameter_array, model_parameter_array
        )
        model_parameter_array_sample = np.append(
            model_parameter_array_sample, 1.02 * model_parameter_array
        ).reshape(3, len(model_parameter_array))

        dependent_predictions = []
        dependent_prediction_errors = []

        for emulator_index in gpe_ordering:
            gp = self.bin_gaussian_process[emulator_index]
            model_values = self.bin_model_values[emulator_index]

            model, errors = gp.predict(
                y=model_values["dependent"],
                t=model_parameter_array_sample,
                return_cov=False,
                return_var=True,
            )

            # Remove fake points required to ensure george returns a prediction.
            dependent_predictions.append(model[1])
            dependent_prediction_errors.append(errors[1])

        return (
            np.array(dependent_predictions),
            np.array(dependent_prediction_errors),
        )

    def predict_values_no_error(
        self, independent: np.array, model_parameters: Dict[str, float]
    ) -> np.array:
        """
        Predict values from the trained emulator contained within this object.
        In cases where the error estimates are not required, this method is
        significantly faster than predict_values().

        Parameters
        ----------

        independent, np.array
            Independent continuous variables to evaluate the emulator
            at. If the emulator is discrete, these are only allowed to be
            the discrete independent variables that the emulator was trained at
            (disregarding the additional 'independent' model parameters, below).
            These can be found in this object in the ``bin_centers`` attribute.

        model_parameters: Dict[str, float]
            The point in model parameter space to create predicted
            values at.

        Returns
        -------

        dependent_predictions, np.array
            Array of predictions, if the emulator is a function f, these
            are the predicted values of f(independent) evaluted at the position
            of the input ``model_parameters``.

        Raises
        ------

        AttributeError
            When the model has not been trained before trying to make a
            prediction, or when attempting to evaluate the model at
            disallowed independent variables.
        """

        if self.bin_gaussian_process is None:
            raise AttributeError(
                "Please train the emulator with fit_model before attempting "
                "to make predictions."
            )

        # First calculate which indices in bin_centers (and hence
        # bin_gaussian_processes) correspond to the requested ``independent``
        # variables.

        array_centers = np.array(self.bin_centers)
        gpe_ordering = []

        for requested_independent_variable in independent:
            try:
                gpe_ordering.append(
                    np.where(array_centers == requested_independent_variable)[0][0]
                )
            except IndexError:
                raise AttributeError(
                    f"Requested independent variable {independent} not valid, ",
                    f"this instance of GPE Bins is only valid at {array_centers}.",
                )

        model_parameter_array = np.array(
            [model_parameters[parameter] for parameter in self.parameter_order]
        )

        # George must predict a value for more than one point at a time, so
        # generate two fake points either side of the one of interest.
        model_parameter_array_sample = np.append(
            0.98 * model_parameter_array, model_parameter_array
        )
        model_parameter_array_sample = np.append(
            model_parameter_array_sample, 1.02 * model_parameter_array
        ).reshape(3, len(model_parameter_array))

        dependent_predictions = []

        for emulator_index in gpe_ordering:
            gp = self.bin_gaussian_process[emulator_index]
            model_values = self.bin_model_values[emulator_index]

            model = gp.predict(
                y=model_values["dependent"],
                t=model_parameter_array_sample,
                return_cov=False,
                return_var=False,
            )

            # Remove fake points required to ensure george returns a prediction.
            dependent_predictions.append(model[1])

        return np.array(dependent_predictions)
